﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace Sociometry
{
    class Sociometer
    {
        public List<Dihotomy> Dihotomies { get; }

        public string[] SocionicType { get; }

        public double[] FeaturesProbability { get; }

        int[] wTim;

        public Sociometer()
        {
            Dihotomies = new List<Dihotomy>
            {
                new Dihotomy("Экстраверсия", "Интроверсия"),
                new Dihotomy("Интуиция", "Сенсорика"),
                new Dihotomy("Логика", "Этика"),
                new Dihotomy("Иррациональность", "Рациональность"),
                new Dihotomy("Беспечность", "Предусмотрительность"),
                new Dihotomy("Уступчивость", "Упрямство"),
                new Dihotomy("Статика", "Динамика"),
                new Dihotomy("Демократия", "Аристократия"),
                new Dihotomy("Тактика", "Стратегия"),
                new Dihotomy("Конструктивизм", "Эмотивизм"),
                new Dihotomy("Позитивизм", "Негативизм"),
                new Dihotomy("Рассудительность", "Решительность"),
                new Dihotomy("Веселость", "Серьезность"),
                new Dihotomy("Процесс", "Результат"),
                new Dihotomy("Квестимность", "Декларативность")
            };

            SocionicType = new string[SocionicTypes.Types.Count];
            int i = 0;
            foreach (var type in SocionicTypes.Types)
            {
                SocionicType[i] = type.Value;
                i++;
            }
            FeaturesProbability = new double[SocionicTypes.Types.Count];
            wTim = new int[16] { 32767, 14520, 8065, 22726, 29203, 13652, 4717, 21802, 25801, 9102,
                                       1207, 17392, 26917, 11874, 2395, 19996 };
        }

        public void ReadParameters(string filename)
        {
            string[] keys = File.ReadAllLines(filename);
            int dihotomiesCount = Dihotomies.Count;
            for (int i = 0; i < dihotomiesCount; i++)
            {
                string firstPolarity = Dihotomies[i].FirstPolarityName;
                string secondPolarity = Dihotomies[i].SecondPolarityName;

                string header = $"[{firstPolarity}|{secondPolarity}]";
                int sectionBegin = Array.IndexOf(keys, header);
                bool sectionExist = (sectionBegin > -1) && (sectionBegin < keys.Length - 1);
                if (sectionExist)
                {
                    for (int j = sectionBegin + 1; j < keys.Length; j++)
                    {
                        if (keys[j].Contains("[") || keys[j].Trim() == "")
                        {
                            break;
                        }
                        else
                        {
                            string value = keys[j];
                            string[] values = value.Split(';');
                            string keyname = values[0];
                            double[] a = new double[] { double.Parse(values[1]), double.Parse(values[4]) };
                            double[] sigma = new double[] { double.Parse(values[2]), double.Parse(values[5]) };
                            double[] factor = new double[] { double.Parse(values[3]), double.Parse(values[6]) };
                            Dihotomies[i].Keys.Add(new Key(keyname, a, sigma, factor));
                        }
                    }
                }
            }
        }

        public void CalcFrequency(string filename)
        {
            int symbolsCount = Tools.CountSymbols(filename);
            string[] words = Tools.WordsFromText(filename);
            string key;
            double keyQuantity;
            double frequency;
            for (int i = 0; i < Dihotomies.Count; i++)
            {
                for (int j = 0; j < Dihotomies[i].Keys.Count; j++)
                {
                    key = Dihotomies[i].Keys[j].Name;
                    if (key == "symbol")
                    {
                        frequency = symbolsCount;
                    }
                    else
                    {
                        keyQuantity = Tools.CountWordQuantity(key, words);
                        frequency = Math.Round(keyQuantity * 10000 / symbolsCount, 2);
                    }
                    Dihotomies[i].Keys[j].Frequency = frequency;
                }
            }
        }

        public void CalcBayes()
        {
            for (int i = 0; i < Dihotomies.Count; i++)
            {
                Dihotomies[i].CalcBayesProbability();
            }
        }


        public void CalcTIM()
        {
            double[] precision = new double[Dihotomies.Count];
            double[] kora = new double[SocionicType.Length];
            double precisionValue;
            int mask = 0;

            for (int i = 0; i < Dihotomies.Count; i++)
            {
                precisionValue = Dihotomies[i].Bayes;
                if (precisionValue < 0.5)
                {
                    precisionValue = 1 - precisionValue;
                }
                else
                {
                    mask ^= (int)Math.Pow(2, i);
                }
                precision[i] = precisionValue;
            }

            int kisk;
            double fullGroup = 0;
            for (int i = 0; i < kora.Length; i++)
            {
                kisk = mask ^ wTim[i];
                kora[i] = 1;
                for (int j = 0; j < precision.Length; j++)
                {
                    int res = kisk & (int)Math.Pow(2, j);
                    if (res == 0)
                    {
                        kora[i] *= precision[j];
                    }
                    else
                    {
                        kora[i] *= (1 - precision[j]);
                    }
                }
                fullGroup += kora[i];
            }

            for (int i = 0; i < kora.Length; i++)
            {
                if (fullGroup != 0)
                    FeaturesProbability[i] = kora[i] / fullGroup;
                else
                    FeaturesProbability[i] = 1 / kora.Length;
            }
        }

        public Dictionary<string, double> GetSortedTypes()
        {
            Dictionary<string, double> sortedTypes = new Dictionary<string, double>();
            string[] types = (string[])SocionicType.Clone();
            double tempProbability;
            string tempSocionicType;
            double[] probabilities = (double[])FeaturesProbability.Clone();
            for (int i = 0; i < probabilities.Length; i++)
            {
                for (int j = i + 1; j < probabilities.Length; j++)
                {
                    if (probabilities[i] < probabilities[j])
                    {
                        tempProbability = probabilities[i];
                        probabilities[i] = probabilities[j];
                        probabilities[j] = tempProbability;

                        tempSocionicType = types[i];
                        types[i] = types[j];
                        types[j] = tempSocionicType;
                    }
                }

            }
            for (int i = 0; i < types.Length; i++)
            {
                sortedTypes.Add(types[i], probabilities[i]);
            }
            return sortedTypes;
        }
    }
}
